#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes to represent the git repository staging area
"""

import itertools
import pathlib
import tempfile
import shutil


class StagingArea(object):
    """
    Class to represent the staging area as an whole.
    It should be used as context manager (with the "with" statement).
    """

    def __init__(self, repository):
        """
        :param repository: GitPython Repository
        :rtype repository: git.repo.base.Repo
        """

        self.temporary_directory = None
        self.files = []  # this will be populated with the temporary files

        if repository.active_branch.is_valid():
            diff = repository.head.commit.diff()
            added_changes = diff.iter_change_type('A')
            modified_changes = diff.iter_change_type('M')
            self.changes = list(itertools.chain(added_changes, modified_changes))
        else:
            # if the branch still doesn't have commits the diff will fail
            self.changes = []

        self.working_dir = pathlib.Path(repository.working_dir)

    def __enter__(self):
        """
        Enter context manager and create a temporary directory
        """
        self.temporary_directory = pathlib.Path(tempfile.mkdtemp())
        self.files = [self.create_temp_file(diff) for diff in self.changes]
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Remove temporary directory on exit
        """
        self.files = []
        shutil.rmtree(str(self.temporary_directory), ignore_errors=True)
        self.temporary_directory = None

    def create_temp_file(self, file_diff):
        """
        Creates a temporary file based on a GitPython file diff
        This method replicates the original project structure and creates a file with the original name

        :param file_diff: Diff for file
        :type file_diff: git.diff.Diff
        """

        if not self.temporary_directory:
            raise ValueError('StagingArea not entered')

        new_version = file_diff.b_blob
        original_relative_path = new_version.path
        content = new_version.data_stream.stream.read()

        temporary_path = self.temporary_directory / original_relative_path
        if not temporary_path.parent.exists():
            temporary_path.parent.mkdir(parents=True)
        with temporary_path.open('wb') as temporary_file:
            temporary_file.write(content)
        return temporary_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Zalando SE

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

from __future__ import print_function, absolute_import

from git import InvalidGitRepositoryError, BadName  # noqa
import git
import os

assert BadName  # silence pyflakes


def get_repository(directory=None):
    """
    git.Repo() fails if it's not in the root of the repository so wee need to try the parents to see if we are inside
    a repository

    :rtype: Optional[git.Repo]
    """
    directory = directory or os.getcwd()
    repo = None
    while repo is None and directory != '/':
        try:
            repo = git.Repo(directory)
        except InvalidGitRepositoryError:
            directory = os.path.dirname(directory)
    return repo



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Zalando SE

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

from __future__ import print_function

import distutils.version as du_version
import pkg_resources


import click
import pip
import requests


PYPI_URL = 'https://pypi.python.org/pypi/{package}/json'


def get_pypi_version(package):
    """
    Gets the latest version from PYPI

    :rtype: distutils.version.LooseVersion
    """
    pypi_data = requests.get(PYPI_URL.format(package=package)).json()
    project_info = pypi_data.get('info', dict())
    pypi_version = project_info.get('version')
    return du_version.LooseVersion(pypi_version)


def upgrade_packages(packages):
    """
    Use pip to upgrade package
    """
    arguments = ['install', '--upgrade']
    arguments.extend(packages)
    pip.main(arguments)


def get_packages():
    """
    Gets a list of packages to be verified.
    """

    for entry_point in pkg_resources.iter_entry_points('turnstile.commands'):
        yield entry_point.dist.version, entry_point.dist.project_name

    for entry_point in pkg_resources.iter_entry_points('turnstile.pre_commit'):
        yield entry_point.dist.version, entry_point.dist.project_name

    for entry_point in pkg_resources.iter_entry_points('turnstile.commit_msg'):
        yield entry_point.dist.version, entry_point.dist.project_name


@click.command('upgrade')
@click.option('--verbose', '-v', is_flag=True)
def cmd(verbose):
    """
    Upgrade Turnstile
    """
    outdated = []
    for package_version, package_name in set(get_packages()):
        pypi_version = get_pypi_version(package=package_name)
        local_version = du_version.LooseVersion(package_version)
        if verbose:
            click.secho('{package_name} {local_version} (PyPi: {pypi_version})'.format(**locals()))
        if local_version < pypi_version:
            outdated.append(package_name)

    if outdated:
        click.secho('The following packages are outdated:', bold=True)
        click.secho('  {packages}'.format(packages=' '.join(outdated)))
        if click.confirm('Do you want to continue?'):
            upgrade_packages(outdated)
    else:
        click.secho('Turnstile is already updated.', bold=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import turnstile.checks as checks
import turnstile.common.output as output


@checks.Check('Not committing to master')
def check(user_configuration, repository_configuration, commit_message):
    """
    Prevents commits to master

    >>> import turnstile.models.message as message

    >>> commit = message.CommitMessage('master', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result = check(None, None, commit)
    >>> result.successful
    False
    >>> result.details
    ['Master branch is protected.']

    :param user_configuration: User specific configuration
    :type user_configuration: git_hooks.common.config.UserConfiguration
    :param repository_configuration: Repository specific configuration
    :type repository_configuration: dict
    :param commit_message:
    :type commit_message: git_hooks.models.message.CommitMessage
    :return: If check passed or not
    :rtype: git_hooks.checks.CheckResult
    """
    logger = output.get_sub_logger('commit-msg', 'protect-master')

    logger.debug('Starting protect-master check...')

    result = checks.CheckResult()
    branch = commit_message.branch
    logger.debug('Branch: %s', branch)

    is_allowed = branch != 'master'
    result.successful = is_allowed
    if not is_allowed:
        result.add_detail("Master branch is protected.")

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Zalando SE

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

from __future__ import absolute_import

import pkg_resources

import click

manager = click.Group()


def get_commands():
    """
    Fetch click commands from extension `turnstile.command` entrypoints.
    The click command functions should be named cmd
    :rtype: Iterator[str, Callable]
    """
    for entry_point in pkg_resources.iter_entry_points('turnstile.commands'):
        try:
            module = entry_point.load()
        except ImportError:
            continue  # ignore broken commands
        command = module.cmd  # type: FunctionType
        yield entry_point.name, command


for name, command in get_commands():
    manager.add_command(command, name=name)


if __name__ == '__main__':
    manager()  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import

import gitconfig
import logging
import pathlib
import pyrsistent
import yaml
import yaml.scanner


logger = logging.getLogger('turnstile.config')


def is_valid_log_verbosity(verbosity):
    """
    >>> is_valid_log_verbosity('WARNING')
    True

    >>> is_valid_log_verbosity('INFO')
    True

    >>> is_valid_log_verbosity('DEBUG')
    True

    >>> is_valid_log_verbosity('SOMETHINGELSE')
    False

    :type verbosity: str|int
    :rtype: bool
    """
    try:
        logging._checkLevel(verbosity)
        return True
    except (ValueError, TypeError):
        return False


def load_repository_configuration(repository_path):
    """
    Load repository specific configurations adding CONFIG_FILE as a parameter with the load file.
    If hooks.config doesn't exist it returns only  {'CONFIG_FILE': 'DEFAULT'}

    :rtype: pyrsistent.pmap
    """
    repository_path = pathlib.Path(repository_path)
    config_path = repository_path / '.turnstile.yml'
    try:
        with config_path.open() as config_file:
            # If the configuration file is empty safe_load returns None and we want a dict
            config = yaml.safe_load(config_file) or {}
            config['CONFIG_FILE'] = config_path
    except IOError:
        config = {'CONFIG_FILE': 'DEFAULT'}
    except TypeError:
        raise ValueError('Invalid Repository Configuration')
    except yaml.scanner.ScannerError as e:
        # Try to give a meaningful error if YAML parsing fails
        context = str(e.context_mark).strip()
        error = '{e.problem} {e.context} {context}'.format(e=e, context=context)
        raise ValueError('Invalid Repository Configuration - {}'.format(error))
    return pyrsistent.pmap(config)


class UserConfiguration(object):
    """
    User specific options


    >>> config=UserConfiguration()
    >>> config.store = dict()  # Don't mess up with real config
    >>> config.verbosity = 'DEBUG'
    >>> config.verbosity
    'DEBUG'
    >>> config.verbosity = 'INFO'
    >>> config.verbosity
    'INFO'
    >>> config.verbosity = 'OTHER'
    Traceback (most recent call last):
        ...
    ValueError: "OTHER" is not a valid verbosity
    >>> config.store['turnstile.verbosity'] = 'BREAKSTUFF'
    >>> config.verbosity
    'INFO'
    """

    def __init__(self, level='global'):
        self.store = gitconfig.GitConfig(level)

    @property
    def verbosity(self):
        """
        Tries to get logging level from git config, falling back to INFO if it is not set or is an invalid value

        :return: Log Level
        :rtype: str
        """
        verbosity = self.store.get('turnstile.verbosity', 'INFO')
        if not is_valid_log_verbosity(verbosity):
            logger.warning('Invalid Verbosity "%s" falling back to INFO', verbosity)
            verbosity = 'INFO'
        return verbosity

    @verbosity.setter
    def verbosity(self, value):
        """
        Store git hook verbosity in git config

        :type value: str | int
        :param value: new verbosity
        """
        if is_valid_log_verbosity(value):
            self.store['turnstile.verbosity'] = value
        else:
            raise ValueError('"{}" is not a valid verbosity'.format(value))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Zalando SE

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

from __future__ import print_function, absolute_import

import sys

import click
import git

import turnstile.checks as checks
import turnstile.common.config as config
import turnstile.common.output as output
import turnstile.models.message as message


@click.command()
@click.argument('message_file_path')
def commit_msg(message_file_path):
    """
    This hook is invoked by git commit, and can be bypassed with --no-verify option. It takes a single parameter,
    the name of the file that holds the proposed commit log message.
    Exiting with non-zero status causes the git commit to abort.

    :param message_file_path: the name of the file that holds the proposed commit log message
    :type message_file_path: string
    """

    repository = git.Repo()
    branch = repository.active_branch.name

    user_configuration = config.UserConfiguration()
    logger = output.get_root_logger('commit-msg')
    logger.setLevel(user_configuration.verbosity)

    logger.debug('Starting Commit-Msg Hook')
    logger.debug('Path to commit message file: %s', message_file_path)

    logger.debug('Repository Working Dir: %s', repository.working_dir)
    logger.debug('Current branch: %s', branch)

    try:
        repository_configuration = config.load_repository_configuration(repository.working_dir)
    except ValueError as e:
        logger.error(str(e))
        raise click.Abort
    logger.debug('Loaded repository configuration: %s', repository_configuration['CONFIG_FILE'])

    logger.debug('Opening commit message file')
    try:
        with open(message_file_path) as message_file:
            str_commit_message = message_file.read()
    except IOError:
        logger.error('Commit message file (%s) not found', message_file_path)
        raise click.Abort
    logger.debug('Commit Message: %s', str_commit_message)

    commit_message = message.CommitMessage(branch, str_commit_message)

    failed_checks = checks.run_checks('commit_msg', user_configuration, repository_configuration, commit_message)

    if failed_checks:
        s = '' if failed_checks == 1 else 's'
        logger.error('%d check%s failed', failed_checks, s)

    sys.exit(failed_checks)

if __name__ == '__main__':
    commit_msg()  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Zalando SE

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

import pkg_resources

import turnstile.common.output as output
# we need to import the following modules so import will work on python3


class CheckIgnore(Exception):
    """
    Exception to tell an check should be ignored in a specific case.
    """
    pass


class CheckResult(object):
    def __init__(self, successful=True, details=None):
        """
        :type successful: bool
        :type details: list
        """
        self.successful = successful
        self.details = details or list()

    def add_detail(self, detail):
        """
        Adds detail to a check
        :type detail: str
        """
        self.details.append(detail)


class Check(object):

    """
    Decorator that wraps a check
    """

    def __init__(self, description):
        """
        :param description: Check description
        :type description: str
        """
        self.description = description

    def __call__(self, function):
        function.description = self.description
        return function


def get_checks(hook_name):
    """
    Load all the checks for the hook

    :param hook_name: Which hook is fetching the checks
    :type hook_name: str
    :rtype: [str, FunctionType]
    """
    logger = output.get_sub_logger(hook_name.replace('_', '-'), 'get_checks')

    group_name = 'turnstile.{}'.format(hook_name)
    for entry_point in pkg_resources.iter_entry_points(group_name):
        try:
            module = entry_point.load()
        except ImportError:  # pragma: no cover
            logger.error('%s not found', entry_point.name)
            continue
        check = module.check  # type: FunctionType
        yield entry_point.name, check


def run_checks(hook_name, user_configuration, repository_configuration, check_object, ):
    """
    Runs checks for hooks

    :param hook_name: Which hook is running the checks
    :type hook_name: str
    :param user_configuration: User specific configuration
    :type user_configuration: git_hooks.common.config.UserConfiguration
    :param repository_configuration: Repository specific configuration
    :type repository_configuration: dict
    :param check_object: Object to check, either a CommitMessage or a StagingArea
    :return: Number of failed checks
    :rtype: int
    """

    logger = output.get_sub_logger(hook_name.replace('_', '-'), 'run_checks')
    failed_checks = 0
    checklist = repository_configuration.get('checks', [])
    check_functions = get_checks(hook_name)
    checks_to_run = (check for check_name, check in check_functions if check_name in checklist)
    for check in checks_to_run:

        try:
            result = check(user_configuration, repository_configuration, check_object)
        except CheckIgnore:
            logger.debug('Check was ignored')
            continue

        if result.successful:
            logger.info('✔ %s', check.description)
            for detail in result.details:
                logger.info('  %s', detail)
        else:
            failed_checks += 1
            logger.error('✘ %s', check.description)
            for detail in result.details:
                logger.error('  %s', detail)
    return failed_checks



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Zalando SE

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""


from __future__ import print_function, absolute_import

import sys

import click
import git

import turnstile.checks as checks
import turnstile.common.config as config
import turnstile.common.output as output
import turnstile.models.staging as staging


@click.command()
def pre_commit():
    """
    This hook is invoked by git commit, and can be bypassed with --no-verify option. It takes no parameter, and is
    invoked before obtaining the proposed commit log message and making a commit.
    Exiting with non-zero status from this script causes the git commit to abort.
    """

    repository = git.Repo()

    user_configuration = config.UserConfiguration()

    logger = output.get_root_logger('pre-commit')
    logger.setLevel(user_configuration.verbosity)
    logger.debug('Starting Pre-Commit Hook')
    logger.debug('Repository Working Dir: %s', repository.working_dir)

    try:
        repository_configuration = config.load_repository_configuration(repository.working_dir)
    except ValueError as e:
        logger.error(e)
        raise click.Abort
    logger.debug('Loaded repository configuration: %s', repository_configuration['CONFIG_FILE'])

    staging_area = staging.StagingArea(repository)
    logger.debug('Changed Files: %d', len(staging_area.changes))

    failed_checks = checks.run_checks('pre_commit', user_configuration, repository_configuration, staging_area)

    if failed_checks:
        s = '' if failed_checks == 1 else 's'
        logger.error('%d check%s failed', failed_checks, s)

    sys.exit(failed_checks)

if __name__ == '__main__':
    pre_commit()  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import turnstile.checks as checks
import turnstile.common.output as output
import re


@checks.Check('Branch name is valid')
def check(user_configuration, repository_configuration, commit_message):
    """
    Check if the branch name matches the allowed pattern. Master is always allowed

    By default this check only allows master
    >>> import turnstile.models.message as message
    >>> commit = message.CommitMessage('master', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result = check(None, {}, commit)
    >>> result.successful, result.details
    (True, [])

    >>> commit = message.CommitMessage('feature/42', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result = check(None, {}, commit)
    >>> result.successful, result.details
    (False, ["feature/42 doesn't match any allowed pattern."])

    But you can add more allowed patterns
    >>> allow_feature_release = {'branch-pattern': {'allowed': ['^feature/', '^release/R']}}
    >>> commit = message.CommitMessage('release/R10', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result = check(None, allow_feature_release, commit)
    >>> result.successful, result.details
    (True, [])

    >>> allow_feature_release = {'branch-pattern': {'allowed': ['^feature/', '^release/R']}}
    >>> commit = message.CommitMessage('release/R10', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result = check(None, allow_feature_release, commit)
    >>> result.successful, result.details
    (True, [])

    >>> allow_feature_release = {'branch-pattern': {'allowed': ['^feature/', '^release/R']}}
    >>> commit = message.CommitMessage('release/broken', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result = check(None, allow_feature_release, commit)
    >>> result.successful, result.details
    (False, ["release/broken doesn't match any allowed pattern."])

    :param user_configuration: User specific configuration
    :type user_configuration: git_hooks.common.config.UserConfiguration
    :param repository_configuration: Repository specific configuration
    :type repository_configuration: dict
    :param commit_message:
    :type commit_message: git_hooks.models.message.CommitMessage
    :return: If check passed or not
    :rtype: git_hooks.checks.CheckResult
    """
    logger = output.get_sub_logger('commit-msg', 'branch-pattern')

    logger.debug('Starting branch-pattern check...')

    result = checks.CheckResult()
    branch = commit_message.branch
    logger.debug('Branch: %s', branch)

    check_options = repository_configuration.get('branch-pattern', {})
    allowed = check_options.get('allowed', [])
    allowed.append('master')  # master is always allowed

    logger.debug('Allowed Patterns: %s', allowed)

    is_allowed = any(re.match(pattern, branch) for pattern in allowed)
    result.successful = is_allowed
    if not is_allowed:
        template = "{branch} doesn't match any allowed pattern."
        result.add_detail(template.format(branch=branch))

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Zalando SE

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

import rfc3986
import re

import turnstile.common.github as github


class Specification(object):
    def __init__(self, identifier, allowed_formats, allowed_uri_schemes):
        """

        :type identifier: str
        :type allowed_formats: Iterator
        :return:
        """
        self.identifier = identifier
        self.allowed_formats = allowed_formats

        # For URIs
        self.allowed_schemes = allowed_uri_schemes or ['https', 'offline']

    @property
    def format(self):
        """
        Checks the identifier against the allowed formats and returns the first that matches or None if none do.

        :rtype: Optional(str)
        """
        validators = {'github': self.validate_github,
                      'jira': self.validate_jira,
                      'uri': self.validate_uri}

        for format in self.allowed_formats:
            if validators[format]():
                return format

    @property
    def valid(self):
        """
        Goes through all allowed format validators and checks if the identifier is valid in at least one of them

        >>> Specification('https://short.url/spec', {'uri'}, ['https', 'offline']).valid
        True
        >>> Specification('https://10[0]0', {'uri'}, ['https', 'offline']).valid
        False
        >>> Specification('noscheme', {'uri'}, ['https', 'offline']).valid
        False
        >>> Specification('#32', {'uri'}, ['https', 'offline']).valid
        False
        >>> Specification('#32', {'uri', 'github'}, ['https', 'offline']).valid
        True
        """

        return bool(self.format)

    def validate_uri(self):
        """
        Specification URIs have to be valid, have a scheme and be absolute
        :rtype: bool
        """
        uri = rfc3986.uri_reference(self.identifier)
        valid_uri_scheme = uri.scheme in self.allowed_schemes
        return uri.is_valid() and valid_uri_scheme and uri.is_absolute()

    def validate_github(self):
        """
        https://help.github.com/articles/writing-on-github/#references

        :rtype: bool

        >>> Specification('#42', {'github'}, []).validate_github()
        True

        >>> Specification('zalando/turnstile#42', {'github'}, []).validate_github()
        True

        >>> Specification('GH-42', {'github'}, []).validate_github()
        True

        >>> Specification('zalando#42', {'github'}, []).validate_github()
        True

        >>> Specification('32', {'github'}, []).validate_github()
        False
        """
        return bool(github.extract_issue_number(self.identifier))

    def validate_jira(self):
        """
        :rtype: bool

        >>> Specification('CD-42', {'jira'}, []).validate_jira()
        True

        >>> Specification('PROJECT-42', {'jira'}, []).validate_jira()
        True

        >>> Specification('#PROJECT-42', {'jira'}, []).validate_jira()
        True

        >>> Specification('#42', {'jira'}, []).validate_jira()
        False

        >>> Specification('invalid', {'jira'}, []).validate_jira()
        False
        """

        regex = r'#?(?P<id>[A-Z]+-[0-9]+)'
        return bool(re.match(regex, self.identifier))

    def __str__(self):
        return self.identifier


def get_specification(commit_message, allowed_formats, allowed_uri_schemes):
    """
    Extracts the specification URI from the commit_message and creates the appropriate specification instance based on
    the URI scheme. If scheme is missing from the URI the default_specification_format will be used.

    >>> spec = get_specification('something', {'uri'}, {'https'})
    >>> spec.valid
    False

    :type commit_message:str
    :type default_specification_format:str
    :type allowed_uri_schemes: List[str]
    :return: BaseSpecification
    """
    try:
        specification_str, _ = commit_message.split(' ', 1)
    except ValueError:
        # If there is only one word (the spec) split will fail
        specification_str = commit_message

    specification = Specification(specification_str, allowed_formats, allowed_uri_schemes)
    return specification



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import

import re

import turnstile.checks as checks
import turnstile.common.output as output


@checks.Check('Branch release is valid')
def check(user_configuration, repository_configuration, commit_message):
    """
    Check if the release of release branches matches a pattern. By default this pattern is ^R(?:\d|\_|\.)+$ but it's
    configurable

    :param user_configuration: User specific configuration
    :type user_configuration: git_hooks.common.config.UserConfiguration
    :param repository_configuration: Repository specific configuration
    :type repository_configuration: dict
    :param commit_message:
    :type commit_message: git_hooks.models.message.CommitMessage
    :return: If check passed or not
    :rtype: git_hooks.checks.CheckResult
    """

    logger = output.get_sub_logger('commit-msg', 'branch-release')
    logger.debug('Starting branch-release check...')

    result = checks.CheckResult()
    branch = commit_message.branch
    logger.debug('Branch: %s', branch)
    if not branch.startswith('release/'):
        logger.debug("%s isn't a release branch, ignoring.", branch)
        raise checks.CheckIgnore

    check_options = repository_configuration.get('branch-release', {})
    pattern = check_options.get('pattern', '^R(?:\d|\_|\.)+$')
    branch_type, release = branch.split('/', 1)

    matches_pattern = bool(re.match(pattern, release))
    result.successful = matches_pattern
    if not matches_pattern:
        template = "'{release}' doesn't match '{pattern}'."
        result.add_detail(template.format(release=release, pattern=pattern))

    return result



import click

import pathlib

import git

import turnstile.common.output as output


def remove_hook(name, path):
    """
    Remove a hook from path

    :type name: str
    :type path: Path
    """

    logger = output.get_sub_logger('manager-remove', 'remove_hook')
    logger.debug('Removing %s hook.', name)
    if not path.exists():
        logger.debug('%s Hook doesn\'t exist.', name)
    elif click.confirm('Are you sure you want to remove {} hook?'.format(name)):
        path.unlink()
        logger.info('Removed %s hook', name)
    else:
        logger.info('Kept %s hook.', name)


@click.command('remove')
@click.option('-v', 'verbosity', flag_value='DEBUG', help='Use Verbose Mode')
@click.option('-n', 'verbosity', flag_value='INFO', default=True, help='Use Normal Verbosity (default)')
@click.option('-q', 'verbosity', flag_value='WARNING', help='Use Quiet Mode')
def cmd(verbosity):
    """
    Remove git hooks from repository
    """

    logger = output.get_root_logger('manager-remove')
    logger.setLevel(verbosity)

    logger.info('Remove Git Hooks')
    try:
        repository = git.Repo()
    except git.InvalidGitRepositoryError:
        logger.error('This command must be executed inside a repository.')
        raise click.Abort

    hook_dir = pathlib.Path(repository.git_dir) / 'hooks'
    pre_commit_path = hook_dir / 'pre-commit'
    commit_msg_path = hook_dir / 'commit-msg'
    remove_hook('Pre-Commit', pre_commit_path)
    remove_hook('Commit-Msg', commit_msg_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import turnstile.checks as checks
import turnstile.common.output as output


@checks.Check('Branch type is valid')
def check(user_configuration, repository_configuration, commit_message):
    """
    Check if the branch type is allowed. The branch type is the prefix of the branch name, for example feature/CD-100 is
    a feature branch.

    By default only master is allowed
    >>> import turnstile.models.message as message
    >>> commit_1 = message.CommitMessage('master', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result_1 = check(None, {}, commit_1)
    >>> result_1.successful, result_1.details
    (True, [])

    >>> commit_2 = message.CommitMessage('feature/ABC', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result_2 = check(None, {}, commit_2)
    >>> result_2.successful, result_2.details
    (False, ["'feature' type is not allowed. Allowed types are: master."])

    But you can configure it
    >>> allow_feature_release = {'branch-type': {'allowed': ['feature', 'release']}}
    >>> commit_3 = message.CommitMessage('feature/ABC', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result_3 = check(None, allow_feature_release, commit_3)
    >>> result_3.successful, result_3.details
    (True, [])

    >>> allow_feature_release = {'branch-type': {'allowed': ['feature', 'release']}}
    >>> commit_4 = message.CommitMessage('other/ABC', 'https://github.com/zalando-bus/turnstile/issues/42 message')
    >>> result_4 = check(None, allow_feature_release, commit_4)
    >>> result_4.successful, result_4.details
    (False, ["'other' type is not allowed. Allowed types are: feature, release, master."])

    :param user_configuration: User specific configuration
    :type user_configuration: git_hooks.common.config.UserConfiguration
    :param repository_configuration: Repository specific configuration
    :type repository_configuration: dict
    :param commit_message:
    :type commit_message: git_hooks.models.message.CommitMessage
    :return: If check passed or not
    :rtype: git_hooks.checks.CheckResult
    """
    logger = output.get_sub_logger('commit-msg', 'branch-type')

    logger.debug('Starting branch-type check...')

    result = checks.CheckResult()
    branch = commit_message.branch
    logger.debug('Branch: %s', branch)

    check_options = repository_configuration.get('branch-type', {})
    allowed = check_options.get('allowed', [])

    logger.debug('Allowed Patterns: %s', allowed)

    # add a / after type name because branches should be TYPE/* and master is always allowed
    is_allowed = branch == 'master' or any(branch.startswith(branch_type + '/') for branch_type in allowed)
    result.successful = is_allowed
    if not is_allowed:
        branch_type = branch.split('/').pop(0)
        allowed.append('master')  # make it clear it can also be master
        template = "'{branch_type}' type is not allowed. Allowed types are: {allowed}."
        result.add_detail(template.format(branch_type=branch_type, allowed=', '.join(allowed)))

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import turnstile.checks as checks
import turnstile.common.output as output
import turnstile.models.specifications as specifications


@checks.Check('Specification is valid')
def check(user_configuration, repository_configuration, commit_message):
    """
    Check if the specification is valid.

    :param user_configuration: User specific configuration
    :type user_configuration: git_hooks.common.config.UserConfiguration
    :param repository_configuration: Repository specific configuration
    :type repository_configuration: dict
    :param commit_message:
    :type commit_message: git_hooks.models.message.CommitMessage
    :return: If check passed or not
    :rtype: git_hooks.checks.CheckResult
    """

    logger = output.get_sub_logger('commit-msg', 'specification')
    logger.debug('Starting specification check...')
    logger.debug('Commit Message: %s', commit_message.message)

    if commit_message.message.startswith('Merge'):
        logger.debug("Commit is a merge, ignoring.")
        raise checks.CheckIgnore

    check_options = repository_configuration.get('specification', {})
    allowed_schemes = check_options.get('allowed_schemes', ['https', 'offline'])
    allowed_formats = check_options.get('allowed_formats', {'uri'})
    logger.debug("Allowed schemes: %s", allowed_schemes)

    result = checks.CheckResult()
    specification = specifications.get_specification(commit_message.message, allowed_formats, allowed_schemes)
    is_valid_uri = specification.valid

    logger.debug('Specification: %s', specification)
    logger.debug("Specification is valid: %s", is_valid_uri)

    result.successful = is_valid_uri
    if not is_valid_uri:
        result.add_detail('{spec} is not a valid specification.'.format(spec=specification))

    return result



import click

import turnstile.common.config as config


@click.command('config')
def cmd():
    """
    Set configuration
    """
    # TODO implement local and global

    user_config = config.UserConfiguration()

    current_verbosity = user_config.verbosity
    verbosity_levels = ['WARNING', 'INFO', 'DEBUG']
    print('Select the git hook verbosity (Current Value {}):'.format(current_verbosity))
    for i, level in enumerate(verbosity_levels, start=1):
        print('  {i}. {level}'.format(**locals()))
    max_option = len(verbosity_levels)
    option_value = -1
    while not (0 < option_value <= max_option):
        option_value = click.prompt('Please enter an option [1-{n}]'.format(n=max_option), default=2)
    verbosity = verbosity_levels[option_value-1]
    user_config.verbosity = verbosity



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Zalando SE

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

from __future__ import print_function

import click
import webbrowser

import turnstile.common.config as config
import turnstile.common.git as git
import turnstile.common.github as github
import turnstile.models.specifications as specifications


@click.command('view')
@click.argument('reference', required=False)
def cmd(reference='HEAD'):
    """
    Opens the specification for commit
    """

    repository = git.get_repository()
    if not repository:
        click.secho('This command must be executed inside a repository.', fg='red', bold=True)
        raise click.Abort

    options = config.load_repository_configuration(repository.working_dir).get('specification', {})
    allowed_schemes = options.get('allowed_schemes', ['https', 'offline'])
    allowed_formats = options.get('allowed_formats', {'uri'})

    try:
        commit = repository.commit(reference)
    except git.BadName:
        click.secho("'{reference}' is not a valid commit reference.".format(reference=reference), fg='red', bold=True)
        raise click.Abort

    specification = specifications.get_specification(commit.message, allowed_formats, allowed_schemes)
    specification_format = specification.format

    if not specification_format:
        click.secho("That commit doesn't have a valid specification.", fg='red', bold=True)
        raise click.Abort

    if specification_format == 'uri':
        url = specification.identifier
    elif specification_format == 'github':
        origin = repository.remote('origin')  # type: git.remote.Remote
        git_url = origin.config_reader.get('url')
        repository = github.extract_repository_from_url(git_url)
        if repository:
            issue = github.extract_issue_number(specification.identifier)
            url = 'https://github.com/{repository}/issues/{issue}'.format(**locals())
        else:
            click.secho("{} is not a github repository.".format(git_url), fg='red', bold=True)
            raise click.Abort
    else:
        url = None

    if url:
        click.secho('Opening {}'.format(url))
        webbrowser.open(url)
    else:
        click.secho("{} specifications aren't supported yet.".format(specification_format), fg='red', bold=True)
        raise click.Abort



"""
Common functions and regex expressions for github
"""

import re

GITHUB_REGEX = re.compile(
    r'^(https://github.com/|\w+@github\.com:)(?P<repository>[a-zA-Z0-9_\-]+/[a-zA-Z0-9_\-]+)\.git$')
GITHUB_ISSUE_REGEX = re.compile(r'^(([a-zA-Z0-9_\-]*|[a-zA-Z0-9_\-]+/[a-zA-Z0-9_\-]+)#|GH-)(?P<issue>\d+)$')


def extract_repository_from_url(git_url):
    """

    :param git_url:
    :return:
    :rtype: str | None


    >>> extract_repository_from_url('https://github.com/zalando/turnstile.git')
    'zalando/turnstile'
    >>> extract_repository_from_url('git@github.com:zalando-stups/senza.git')
    'zalando-stups/senza'
    >>> extract_repository_from_url('git@bitbucket.org:jmcs/somerepo.git') is None
    True
    """

    match = GITHUB_REGEX.match(git_url)
    if match:
        return match.group('repository')
    else:
        return None


def extract_issue_number(github_reference):
    """
    Extracts the issue number from the github reference

    :param github_reference:
    :type github_reference: str
    :return:
    :rtype: int | None

    >>> extract_issue_number('#42')
    42
    >>> extract_issue_number('GH-24')
    24
    >>> extract_issue_number('zalando#26')
    26
    >>> extract_issue_number('zalando/turnstile#36')
    36
    >>> extract_issue_number('JIRA-1000') is None
    True
    """

    match = GITHUB_ISSUE_REGEX.match(github_reference)
    if match:
        return int(match.group('issue'))
    else:
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Zalando SE

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

from __future__ import print_function

import click
import sys

import turnstile.common.git as git
import turnstile.common.config as config
import turnstile.models.specifications as specifications


@click.command('specification')
@click.option('--verbose', '-v', is_flag=True)
@click.argument('revision_range', required=False)
def cmd(verbose, revision_range):
    """
    Verifies if the commit messages in a range of revisions have valid specifications.

    This command takes the same revision ranges as ``git log`` to specify which commits are processed

    When using the verbose mode merge commits are printed otherwise they are simply ignored
    """

    repository = git.get_repository()
    if not repository:
        click.secho('This command must be executed inside a repository.', fg='red', bold=True)
        raise click.Abort
    commits = list(repository.iter_commits(revision_range))
    invalid = 0
    options = config.load_repository_configuration(repository.working_dir).get('specification', {})
    allowed_schemes = options.get('allowed_schemes', ['https', 'offline'])
    allowed_formats = options.get('allowed_formats', {'uri'})

    for commit in commits:
        is_a_merge = len(commit.parents) > 1
        if is_a_merge and not verbose:
            continue
        short_hash = commit.hexsha[:7]
        first_line = commit.message.splitlines()[0]
        # TODO make configurable
        specification = specifications.get_specification(commit.message, allowed_formats, allowed_schemes)
        if specification.valid:
            click.secho(' ✔ ', bg='green', fg='white', nl=False)
        elif is_a_merge:
            click.secho('   ', fg='yellow', nl=False)
        else:
            invalid += 1
            click.secho(' ✘ ', bg='red', fg='white', nl=False)
        click.secho(' {} '.format(short_hash), fg='white' if is_a_merge else 'yellow', nl=False, dim=is_a_merge)
        click.secho(first_line, dim=is_a_merge)

    if invalid:
        if invalid == 1:
            message = '1 commit has invalid specification.'
        else:
            message = '{n} commits have invalid specifications.'.format(n=invalid)
        click.secho(message, fg='red', bold=True)

    sys.exit(invalid)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging


def get_root_logger(command):
    output = logging.getLogger('turnstile.'+command)
    output.addHandler(logging.StreamHandler())
    output.setLevel(logging.ERROR)
    return output


def get_sub_logger(parent, name):
    return logging.getLogger('turnstile.{}.{}'.format(parent, name))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

from turnstile.version import version


class PyTest(TestCommand):
    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.cov = None
        self.pytest_args = ['--cov', 'turnstile',
                            '--cov-report', 'term-missing',
                            '--doctest-modules', 'turnstile', 'tests']

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name='turnstile-core',
    packages=find_packages(),
    version=version,
    description='Turnstile - Zalando Local Git Hooks',
    author='Zalando SE',
    url='https://github.com/zalando/turnstile',
    license='Apache License Version 2.0',
    install_requires=['click', 'GitPython', 'pathlib', 'PyYAML', 'rfc3986', 'MapGitConfig', 'requests', 'pip',
                      'pyrsistent'],
    tests_require=['pytest-cov', 'pytest', 'pytest-mock', ],
    cmdclass={'test': PyTest},
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Version Control',
    ],
    long_description='Turnstile - Zalando Local Git Hooks',
    entry_points={'console_scripts': ['turnstile = turnstile.manager:manager',
                                      'zalando-turnstile-commit-msg = turnstile.commit_msg:commit_msg',
                                      'zalando-turnstile-pre-commit = turnstile.pre_commit:pre_commit'],
                  'turnstile.commands': ['config = turnstile.manager_subcommands.config',
                                         'install = turnstile.manager_subcommands.install',
                                         'open-spec = turnstile.manager_subcommands.open_spec',
                                         'remove = turnstile.manager_subcommands.remove',
                                         'specification = turnstile.manager_subcommands.specification',
                                         'upgrade = turnstile.manager_subcommands.upgrade',
                                         'version = turnstile.manager_subcommands.version'],
                  'turnstile.commit_msg': ['branch_pattern = turnstile.checks.commit_msg.branch_pattern',
                                           'branch_release = turnstile.checks.commit_msg.branch_release',
                                           'branch_type = turnstile.checks.commit_msg.branch_type',
                                           'protect_master = turnstile.checks.commit_msg.protect_master',
                                           'specification = turnstile.checks.commit_msg.specification']}
)



import click

import pathlib

import git

import turnstile.common.output as output


def install_hook(name, path, wrapper_command):
    """
    Installs a hook in path

    :type name: str
    :type path: Path
    :type wrapper_command: str
    """
    logger = output.get_sub_logger('manager-install', 'install_hook')
    logger.debug('Installing %s hook.', name)
    if not path.exists() or click.confirm('{} hook already exists. Do you want to overwrite it?'.format(name)):
        with path.open('wb+') as pre_commit_hook:
            pre_commit_hook.write(wrapper_command.encode('utf-8'))
        path.chmod(0o755)  # -rwxr-xr-x
        logger.info('Installed %s hook', name)
    else:
        logger.info('Skipped %s hook installation.', name)


@click.command('install')
@click.option('-v', 'verbosity', flag_value='DEBUG', help='Use Verbose Mode')
@click.option('-n', 'verbosity', flag_value='INFO', default=True, help='Use Normal Verbosity (default)')
@click.option('-q', 'verbosity', flag_value='WARNING', help='Use Quiet Mode')
def cmd(verbosity):
    """
    Install git hooks in repository
    """

    logger = output.get_root_logger('manager-install')
    logger.setLevel(verbosity)

    logger.info('Installing Git Hooks')
    try:
        repository = git.Repo()
    except git.InvalidGitRepositoryError:
        logger.error('This command must be executed inside a repository.')
        raise click.Abort

    hook_dir = pathlib.Path(repository.git_dir) / 'hooks'
    pre_commit_path = hook_dir / 'pre-commit'
    commit_msg_path = hook_dir / 'commit-msg'
    install_hook('Pre-Commit', pre_commit_path, 'zalando-turnstile-pre-commit')
    install_hook('Commit-Msg', commit_msg_path, 'zalando-turnstile-commit-msg $1')
